// Copyright (c) 2016, root.ext@gmail.com. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

part of dna;

class DynamicLibrary<T> {
  final int STACK_SIZE = 256;
  static List<int> sizes = getSizes();

  Type type;
  int libraryPointer;
  Function resolveMethodFunction;

  DynamicLibrary([int resolveMethodFunction(DynamicLibrary that, Invocation invocation)]){
    this.type = T;
    assert(type != dynamic);
    //print('type $type');

    this.resolveMethodFunction = resolveMethodFunction;

    var libTag = Util.classAnnot(reflectClass(type), Library);
    assert(libTag != null);
    var name = libTag.getField(#name).reflectee;
    var altName = libTag.getField(#altName).reflectee;
    assert(name != null);

    this.libraryPointer = loadLibrary(name);
    if (this.libraryPointer == 0 && altName != null){
      this.libraryPointer = loadLibrary(altName);
    }
    //assert(libraryPointer != 0);
    //print('lib pointer $libraryPointer');
  }

  dynamic noSuchMethod(Invocation invocation){
    //print('method ${MirrorSystem.getName(invocation.memberName)}');

    var classMirror = reflectClass(type);
    var methods = classMirror.instanceMembers.values
        .where((e)=> e.isRegularMethod && e.owner == classMirror && e.simpleName == invocation.memberName).toList();
    assert(methods.isNotEmpty);
    var methodMirror = methods.first;

    //Array with TypedData object for params
    var paramsData = new List<TypedData>(invocation.positionalArguments.length);
    var methodPointer = resolveMethod(invocation);

    var vm = createCallVM(STACK_SIZE);
    assert(vm != 0);

    before(invocation, methodMirror, vm, paramsData);
    var result = call(methodMirror, vm, methodPointer);
    after(invocation, methodMirror, paramsData);

    freeCallVM(vm);
    return result;
  }

  int resolveMethod(Invocation invocation){
    if (this.resolveMethodFunction != null){
      var pointer = this.resolveMethodFunction(this, invocation);
      return pointer;
    }

    var methodPointer = findSymbol(this.libraryPointer, MirrorSystem.getName(invocation.memberName));
    assert(methodPointer != 0);
    //print('method pointer $methodPointer');
    return methodPointer;
  }

  void iterateParams(Invocation invocation, MethodMirror methodMirror,
    processParam(dynamic param, Type type, ParameterMirror paramMirror, int index)){

    var paramMirrors = methodMirror.parameters;

    int index = 0;
    for (var paramMirror in paramMirrors) {
      //print('param ${MirrorSystem.getName(param.simpleName)}');
      var paramTag = Util.paramAnnot(paramMirror, Param);
      var paramType = paramTag != null ? paramTag.getField(#type).reflectee : dartTypeToCtype(paramMirror.type);
      var param = invocation.positionalArguments[index];

      processParam(param, paramType, paramMirror, index);
      index++;
    }
  }

  void before(Invocation invocation, MethodMirror methodMirror, int vm, List<TypedData> data) {
    var nativeParams = new List<Object>();
    var signature = new List<String>();

    iterateParams(invocation, methodMirror, (param, type, mirror, index){
      signature.add(new String.fromCharCodes([type]));
      nativeParams.add(param);

      switch (type){
        case C_POINTER: {

        }
        break;
        case C_STRING: {
          var outTag = Util.paramAnnot(mirror, Out);
          if (outTag != null){
            String str = param.value;
            data[index] = new Uint8List(str.length);
            nativeParams[nativeParams.length-1] = data[index];
            signature[signature.length-1] = new String.fromCharCodes([TYPEDDATA]);
          } else {
            nativeParams[nativeParams.length-1] = new Uint8List.fromList(param.codeUnits);
            signature[signature.length-1] = new String.fromCharCodes([TYPEDDATA]);
          }
        }
        break;
        case LISTINT: {
          var outTag = Util.paramAnnot(mirror, Out);
          if (outTag != null){
            List<int> list = param.value;
            data[index] = new Uint32List(list.length);
            nativeParams[nativeParams.length-1] = data[index];
            signature[signature.length-1] = new String.fromCharCodes([TYPEDDATA]);
          }
        }
        break;
        case LISTSTRING: {
          var outTag = Util.paramAnnot(mirror, Out);
          if (outTag != null) {
            throw new UnimplementedError();
          }
        }
        break;
        case C_STRUCT: {
          var outTag = Util.paramAnnot(mirror, Out);
          if (outTag != null) {
            data[index] = encodeStruct(param.value);
          } else {
            data[index] = encodeStruct(param);
          }

          nativeParams[nativeParams.length-1] = data[index];
          signature[signature.length-1] = new String.fromCharCodes([TYPEDDATA]);
        }
        break;
      }
    });

    pushParameters(vm, signature.join(''), nativeParams);
  }

  void after(Invocation invocation, MethodMirror methodMirror, List<TypedData> data) {

    iterateParams(invocation, methodMirror, (param, type, mirror, index) {
      switch (type) {
        case C_POINTER: {

        }
        break;
        case C_STRING: {
          var outTag = Util.paramAnnot(mirror, Out);
          if (outTag != null) {
            Uint8List list = data[index];
            param.value = new String.fromCharCodes(list);
          }
        }
        break;
        case LISTINT: {
          var outTag = Util.paramAnnot(mirror, Out);
          if (outTag != null){
            Uint32List t = data[index];
            param.value = t.toList();
          }
        }
        break;
        case LISTSTRING: {
          var outTag = Util.paramAnnot(mirror, Out);
          if (outTag != null) {
            throw new UnimplementedError();
          }
        }
        break;
        case C_STRUCT: {
          var outTag = Util.paramAnnot(mirror, Out);
          if (outTag != null) {
            param.value = decodeStruct(data[index], param.value);
          }
        }
        break;
      }
    });
  }

  dynamic call(MethodMirror methodMirror, int vm, int methodPointer){

    var retTag = Util.methodAnnot(methodMirror, Method);
    var retType = retTag != null ? retTag.getField(#type).reflectee : dartTypeToCtype(methodMirror.returnType);

    var result;
    if (retType == VOID){
      callFunction(vm, methodPointer, new String.fromCharCodes([retType]));
    } else {
      result = callFunction(vm, methodPointer, new String.fromCharCodes([retType]));
    }

    return result;
  }

  static int dartTypeToCtype(Object type){
    if (type == reflectClass(int)){
      return C_INT;
    } else if (type == reflectClass(double)){
      return C_DOUBLE;
    } else if (type == reflectClass(bool)){
      return C_BOOL;
    } else if (type == reflectClass(String)){
      return C_STRING;
    } else {
      return VOID;
    }
  }

  static int ctypeToSize(int type){
    switch (type){
      case C_CHAR:
      case C_UCHAR:
        return sizes[0];
      case C_SHORT:
      case C_USHORT:
        return sizes[1];
      case C_INT:
      case C_UINT:
        return sizes[2];
      case C_LONG:
      case C_ULONG:
        return sizes[3];
      case C_LONGLONG:
      case C_ULONGLONG:
        return sizes[4];
      case C_BOOL:
        return sizes[5];
      case C_FLOAT:
        return sizes[6];
      case C_DOUBLE:
        return sizes[7];
      case C_POINTER:
        return sizes[8];
      default:
        return 0;
    }
  }
}


